
#include "gmain.h"
#include "script.h"
#include "resource.h"


#define ITEM_LANGUAGE(id, str, def)                     \
{                                                       \
  LANGUAGE_STRING(str, def, Str_Tmp, 1024);             \
  SetDlgItemText(hDlg, id, Str_Tmp);                    \
}


HBITMAP __stdcall Create_Bitmap_From_Icon(HICON hIcon);
BOOL    __stdcall Delete_Key(HKEY hBaseKey, LPCSTR pKeyName);


HBITMAP __stdcall Load_Small_Icon(HINSTANCE hInstance, LPCSTR pResIcon)
{
  HICON            hIcon;

  hIcon = (HICON)LoadImage(hInstance, pResIcon, IMAGE_ICON, 16, 16, 0);
  if(!hIcon) return NULL;

  return Create_Bitmap_From_Icon(hIcon);
}


BOOL __stdcall Get_Type_Info(LPCSTR lpExt, LPSTR lpBuf, DWORD dwBufSize, HBITMAP *phIcon)
{
  HKEY             hKey1, hKey2, hKey3;
  DWORD            dwSize, dwType;
  char             pName[MAX_PATH], pIconFile[MAX_PATH];
  LPSTR            pNum;
  int              iNum;
  HICON            hIconL, hIcon;
  BOOL             bRet = FALSE;

  hIcon   = NULL;
  *phIcon = NULL;

  if(!RegOpenKeyEx(HKEY_CLASSES_ROOT, lpExt, 0, KEY_READ, &hKey1))
  {
    dwSize = MAX_PATH;
    if(!RegQueryValueEx(hKey1, "", NULL, &dwType, (LPBYTE)pName, &dwSize))
    {
      if(!RegOpenKeyEx(HKEY_CLASSES_ROOT, pName, 0, KEY_READ, &hKey2))
      {
        dwSize = dwBufSize;
        if(!RegQueryValueEx(hKey2, "", NULL, &dwType, (LPBYTE)lpBuf, &dwSize))
        {
          bRet = TRUE;
          if(!RegOpenKeyEx(hKey2, "DefaultIcon", 0, KEY_READ, &hKey3))
          {
            dwSize = MAX_PATH;
            if(!RegQueryValueEx(hKey3, "", NULL, &dwType, (LPBYTE)pIconFile, &dwSize))
            {
              pNum = strrchr(pIconFile, ',');
              if(pNum)
              {
                *(pNum++) = 0;
                Clear_String_Spaces(pIconFile);

                if(!(Str2Int(pNum, (LPDWORD)&iNum) & 0xFF00))
                {
                  GetModuleFileName(ghInstance, pName, MAX_PATH);

                  if(!strcmpi(pName, pIconFile))
                  {
                    hIcon = (HICON)LoadImage(ghInstance,
                      MAKEINTRESOURCE(iNum) + IDI_GENS01, IMAGE_ICON, 16, 16, 0);
                  }
                  else
                  {
                    ExtractIconEx(pIconFile, iNum, &hIconL, &hIcon, 1);
                    if(!hIcon)
                    { DestroyIcon(hIcon); hIcon = hIconL; }
                    else if(hIconL) DestroyIcon(hIconL);
                  }
                }
              }
            }
            RegCloseKey(hKey3);
          }
        }
        RegCloseKey(hKey2);
      }
    }
    RegCloseKey(hKey1);
  }

  if(hIcon) *phIcon = Create_Bitmap_From_Icon(hIcon);

  return bRet;
}


BOOL __stdcall Set_Type_Info(LPCSTR pExt, LPCSTR pDesc, int iIconID)
{
  HKEY  hKey, hSubKey;
  DWORD dwDisp;
  char  pBuf[MAX_PATH + 16];
  char  gPath[MAX_PATH];
  int   len;

  if(!pDesc) return FALSE;

  if(RegCreateKeyEx(HKEY_CLASSES_ROOT, pExt, 0, NULL, 0, KEY_WRITE, NULL, &hKey, &dwDisp))
    return FALSE;
  len = sprintf(pBuf, "Gens%s.Type", pExt) + 1;
  RegSetValueEx(hKey, "", 0, REG_SZ, (LPBYTE)pBuf, len);
  RegCloseKey(hKey);

  if(RegCreateKeyEx(HKEY_CLASSES_ROOT, pBuf, 0, NULL, 0, KEY_WRITE, NULL, &hKey, &dwDisp))
    return FALSE;
  RegSetValueEx(hKey, "", 0, REG_SZ, (LPBYTE)pDesc, strlen(pDesc) + 1);

  GetModuleFileName(ghInstance, gPath, MAX_PATH);

  if(iIconID >= IDI_GENS01 && iIconID <= IDI_GENS07)
  {
    if(RegCreateKeyEx(hKey, "DefaultIcon", 0, NULL, 0, KEY_WRITE, NULL, &hSubKey, &dwDisp))
    { RegCloseKey(hKey); return FALSE; }
    len = sprintf(pBuf, "%s,%i", gPath, iIconID - IDI_GENS01) + 1;
    RegSetValueEx(hSubKey, "", 0, REG_SZ, (LPBYTE)pBuf, len);
    RegCloseKey(hSubKey);
  }

  if(RegCreateKeyEx(hKey, "shell\\open\\command", 0, NULL, 0, KEY_WRITE, NULL, &hSubKey, &dwDisp))
    return FALSE;
  len = sprintf(pBuf, "\"%s\" \"%%1\"", gPath);
  RegSetValueEx(hSubKey, "", 0, REG_SZ, (LPBYTE)pBuf, len);
  RegCloseKey(hSubKey);

  RegCloseKey(hKey);

  return TRUE;
}

BOOL __stdcall Remove_Type_Info(LPCSTR pExt)
{
  HKEY  hKey;
  char  pName[128];
  DWORD dwSize, dwType;

  if(!RegOpenKeyEx(HKEY_CLASSES_ROOT, pExt, 0, KEY_READ, &hKey))
  {
    dwSize = sizeof(pName);
    if(!RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)pName, &dwSize))
    {
      Delete_Key(HKEY_CLASSES_ROOT, pName);
    }
    RegCloseKey(hKey);
    Delete_Key(HKEY_CLASSES_ROOT, pExt);
  }

  return TRUE;
}


LRESULT CALLBACK ShellProc(HWND hDlg, UINT uMes, WPARAM wParam, LPARAM lParam)
{
  DWORD dwID;

  switch(uMes)
  {
    case WM_INITDIALOG:
      char    pBuf[128];
      HBITMAP hIcon;

      ITEM_LANGUAGE(IDC_GROUP_ASSOCIATE_TYPES, "Associated file types", "Associated file types");
      if(Get_Type_Info(".ZSG", pBuf, sizeof(pBuf), &hIcon))
      { 
        SetDlgItemText(hDlg, IDC_EDIT_ZSGDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_ZSGTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        if(hIcon)
          SendDlgItemMessage(hDlg, IDC_STATIC_ZSGICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ZSGICON), GWL_USERDATA, 0);
      }
      else
      { 
        ITEM_LANGUAGE(IDC_EDIT_ZSGDESC, "ZSG Type", "Gens Zip Compressed Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_ZSGICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS07)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ZSGICON), GWL_USERDATA, IDI_GENS07);
      }
      if(Get_Type_Info(".SMD", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_SMDDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_SMDTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_SMDICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMDICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_SMDDESC, "SMD Type", "Sega Genesis Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_SMDICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS05)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMDICON), GWL_USERDATA, IDI_GENS05);
      }
      if(Get_Type_Info(".BIN", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_BINDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_BINTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_BINICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_BINICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_BINDESC, "BIN Type", "Sega Genesis Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_BINICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS06)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_BINICON), GWL_USERDATA, IDI_GENS06);
      }
      if(Get_Type_Info(".GEN", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_GENDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_GENTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_GENICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GENICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_GENDESC, "GEN Type", "Sega Genesis Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_GENICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS06)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GENICON), GWL_USERDATA, IDI_GENS06);
      }
      if(Get_Type_Info(".32X", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_32XDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_32XTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_32XICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_32XICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_32XDESC, "32X Type", "Sega 32X Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_32XICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS04)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_32XICON), GWL_USERDATA, IDI_GENS04);
      }
      if(Get_Type_Info(".ISO", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_ISODESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_ISOTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_ISOICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ISOICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_ISODESC, "ISO Type", "SegaCD ISO Image");
        SendDlgItemMessage(hDlg, IDC_STATIC_ISOICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS07)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ISOICON), GWL_USERDATA, IDI_GENS07);
      }
      if(Get_Type_Info(".RAW", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_RAWDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_RAWTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_RAWICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_RAWICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_RAWDESC, "RAW Type", "SegaCD RAW Image");
        SendDlgItemMessage(hDlg, IDC_STATIC_RAWICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS07)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_RAWICON), GWL_USERDATA, IDI_GENS07);
      }
      if(Get_Type_Info(".SMS", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_SMSDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_SMSTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_SMSICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMSICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_SMSDESC, "SMS Type", "Sega Master System Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_SMSICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS07)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMSICON), GWL_USERDATA, IDI_GENS07);
      }
      if(Get_Type_Info(".GG", pBuf, sizeof(pBuf), &hIcon))
      {
        SetDlgItemText(hDlg, IDC_EDIT_GGDESC, pBuf);
        SendDlgItemMessage(hDlg, IDC_CHECK_GGTYPE, BM_SETCHECK, BST_INDETERMINATE, 0);
        SendDlgItemMessage(hDlg, IDC_STATIC_GGICON, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hIcon);
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GGICON), GWL_USERDATA, 0);
      }
      else
      {
        ITEM_LANGUAGE(IDC_EDIT_GGDESC, "GG Type", "Sega Game Gear Rom");
        SendDlgItemMessage(hDlg, IDC_STATIC_GGICON, STM_SETIMAGE, IMAGE_BITMAP,
          (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(IDI_GENS07)));
        SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GGICON), GWL_USERDATA, IDI_GENS07);
      }

      SendDlgItemMessage(hDlg, IDC_EDIT_ZSGDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_SMDDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_BINDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_GENDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_32XDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_ISODESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_RAWDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_SMSDESC, EM_LIMITTEXT, 128, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_GGDESC,  EM_LIMITTEXT, 128, 0);
      break;

    case WM_COMMAND:
      dwID = LOWORD(wParam);
      switch(dwID)
      {
        case IDC_CHECK_ZSGTYPE:
        case IDC_CHECK_SMDTYPE:
        case IDC_CHECK_BINTYPE:
        case IDC_CHECK_GENTYPE:
        case IDC_CHECK_32XTYPE:
        case IDC_CHECK_ISOTYPE:
        case IDC_CHECK_RAWTYPE:
        case IDC_CHECK_SMSTYPE:
        case IDC_CHECK_GGTYPE:
          SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case IDC_EDIT_ZSGDESC:
        case IDC_EDIT_SMDDESC:
        case IDC_EDIT_BINDESC:
        case IDC_EDIT_GENDESC:
        case IDC_EDIT_32XDESC:
        case IDC_EDIT_ISODESC:
        case IDC_EDIT_RAWDESC:
        case IDC_EDIT_SMSDESC:
        case IDC_EDIT_GGDESC:
          if(HIWORD(wParam) == EN_CHANGE)
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case IDC_STATIC_ZSGICON:
        case IDC_STATIC_SMDICON:
        case IDC_STATIC_BINICON:
        case IDC_STATIC_GENICON:
        case IDC_STATIC_32XICON:
        case IDC_STATIC_ISOICON:
        case IDC_STATIC_RAWICON:
        case IDC_STATIC_SMSICON:
        case IDC_STATIC_GGICON:
          if(HIWORD(wParam) == STN_CLICKED)
          {
            int i;
            i = GetWindowLong(GetDlgItem(hDlg, dwID), GWL_USERDATA) + 1;
            if(i < IDI_GENS01 || i > IDI_GENS07) i = IDI_GENS01;
            SetWindowLong(GetDlgItem(hDlg, dwID), GWL_USERDATA, i);
            SendDlgItemMessage(hDlg, dwID, STM_SETIMAGE, IMAGE_BITMAP,
              (LPARAM)Load_Small_Icon(ghInstance, MAKEINTRESOURCE(i)));
            SendDlgItemMessage(hDlg, dwID - 40, BM_SETCHECK, BST_CHECKED, 0);
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          }
          break;
      }
      break;

    case WM_NOTIFY:
      if(((LPNMHDR)lParam)->code == PSN_APPLY)
      {
        char    pBuf[128];
        HRESULT check;

        check = SendDlgItemMessage(hDlg, IDC_CHECK_ZSGTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_ZSGDESC, pBuf, 128);
          Set_Type_Info(".ZSG", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ZSGICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".ZSG");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_SMDTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_SMDDESC, pBuf, 128);
          Set_Type_Info(".SMD", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMDICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".SMD");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_BINTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_BINDESC, pBuf, 128);
          Set_Type_Info(".BIN", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_BINICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".BIN");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_GENTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_GENDESC, pBuf, 128);
          Set_Type_Info(".GEN", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GENICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".GEN");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_32XTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_32XDESC, pBuf, 128);
          Set_Type_Info(".32X", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_32XICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".32X");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_ISOTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_ISODESC, pBuf, 128);
          Set_Type_Info(".ISO", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_ISOICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".ISO");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_RAWTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_RAWDESC, pBuf, 128);
          Set_Type_Info(".RAW", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_RAWICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".RAW");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_SMSTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_SMSDESC, pBuf, 128);
          Set_Type_Info(".SMS", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_SMSICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".SMS");

        check = SendDlgItemMessage(hDlg, IDC_CHECK_GGTYPE, BM_GETCHECK, 0, 0);
        if(check == BST_CHECKED)
        {
          GetDlgItemText(hDlg, IDC_EDIT_GGDESC, pBuf, 128);
          Set_Type_Info(".GG", pBuf,
            GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_GGICON), GWL_USERDATA));
        }
        else if(check == BST_UNCHECKED) Remove_Type_Info(".GG");
      }
      break;
  }
  return 0;
}


HBITMAP __stdcall Create_Bitmap_From_Icon(HICON hIcon)
{
  HBITMAP          hBitmap;
  HDC              hDC;
  BITMAPINFOHEADER bih;
  LPVOID           pBits;

  hDC = CreateCompatibleDC(NULL);
  if(!hDC) { DestroyIcon(hIcon); return NULL; }
  memset(&bih, 0, sizeof(BITMAPINFOHEADER));
  bih.biSize     = sizeof(BITMAPINFOHEADER);
  bih.biPlanes   = 1;
  bih.biBitCount = 24;
  bih.biWidth    = 16;
  bih.biHeight   = 16;
  hBitmap = CreateDIBSection(hDC, (LPBITMAPINFO)&bih, DIB_RGB_COLORS, &pBits, NULL, 0);
  if(!hBitmap) { DestroyIcon(hIcon); DeleteDC(hDC); return NULL; }
  SelectObject(hDC, hBitmap);
  DrawIconEx(hDC, 0, 0, hIcon, 16, 16, 0, (HBRUSH)(COLOR_BTNFACE + 1), DI_NORMAL);
  DeleteDC(hDC);
  DestroyIcon(hIcon);

  return hBitmap;
}


typedef struct _GREG_KEY {
  HKEY              hKey, hOwner;
  char              pName[128];
} GREG_KEY, *PGREG_KEY;

typedef struct _GREG_KEYS {
  int               nKeys;
  PGREG_KEY         pKeys;
} GREG_KEYS, *PGREG_KEYS;


BOOL __stdcall Delete_Key(HKEY hMainKey, LPCSTR pKeyName)
{
  GREG_KEYS gKeys;
  int       n, i;
  char      pName[128];
  DWORD     dwSize;
  LPVOID    ptr;
  FILETIME  fTime;

  gKeys.nKeys = 1;
  gKeys.pKeys = (PGREG_KEY)malloc(sizeof(GREG_KEY));
  if(!gKeys.pKeys) return FALSE;
  gKeys.pKeys[0].hOwner = hMainKey;
  strcpy(gKeys.pKeys[0].pName, pKeyName);

  for(n = 0; n < gKeys.nKeys; n++)
  {
    if(!RegOpenKeyEx(gKeys.pKeys[n].hOwner, gKeys.pKeys[n].pName, 0, KEY_READ, &gKeys.pKeys[n].hKey))
    {
      i      = 0;
      dwSize = 128;
      while(!RegEnumKeyEx(gKeys.pKeys[n].hKey, i++, pName, &dwSize, NULL, NULL, NULL, &fTime))
      {
        ptr = realloc(gKeys.pKeys, (gKeys.nKeys + 1) * sizeof(GREG_KEY));
        if(ptr)
        {
          gKeys.pKeys = (PGREG_KEY)ptr;
          strcpy(gKeys.pKeys[gKeys.nKeys].pName, pName);
          gKeys.pKeys[gKeys.nKeys].hOwner = gKeys.pKeys[n].hKey;
          gKeys.nKeys++;
        }
        dwSize = 128;
      }
    }
  }

  for(n = gKeys.nKeys - 1; n >= 0; n--)
  {
    RegCloseKey(gKeys.pKeys[n].hKey);
    RegDeleteKey(gKeys.pKeys[n].hOwner, gKeys.pKeys[n].pName);
  }

  return TRUE;
}
