
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include "grom.h"
#include "gutil.h"
#include "gdsound.h"
#include "gmain.h"
#include "gmenu.h"
#include "gens.h"
#include "ggenie.h"
#include "cpu_68k.h"
#include "cd_sys.h"
#include "mem_M68K.h"
#include "mem_SH2.h"
#include "vdp_io.h"
#include "state.h"
#include "ccnet.h"
#include "misc.h"
#include "zlib.h"
#include "unzip.h"
#include "wave.h"
#include "script.h"
#include "gsr.h"
#include "vgm.h"
#include "sms.h"
#include "movie.h"
#include "resource.h"

int       File_Type_Index;
GROM_INFO Rom_Info;
char      Rom_Name[G_MAX_ROM_NAME];

LPCSTR      Open_Filter = NULL;

void Get_Name_From_Path(char *Full_Path, char *Name)
{
  int i = 0;

  i = strlen(Full_Path) - 1;

  while ((i >= 0) && (Full_Path[i] != '\\')) i--;

  if(i <= 0)
  {
    Name[0] = 0;
  }
  else
  {
    strcpy(Name, &Full_Path[++i]);
  }
}


void Get_Dir_From_Path(char *Full_Path, char *Dir)
{
  int i = 0;

  i = strlen(Full_Path) - 1;

  while ((i >= 0) && (Full_Path[i] != '\\')) i--;

  if(i <= 0)
  {
    Dir[0] = 0;
  }
  else
  {
    strncpy(Dir, Full_Path, ++i);
    Dir[i] = 0;
  }
}


void __stdcall Update_Recent_Rom(LPCSTR pszRom, int iType)
{
  int   i, exist;
  char  szTmp[GMAX_PATH];

  if(pszRom)
  {
    for(exist = i = 0; i < 9; i++)
    {
      if(!*Recent_Rom[i]) break;
      else if(!strcmpi(Recent_Rom[i], pszRom))
      { exist = 1; break; }
    }

    if(exist)
    {
      if(i)
      {
        strcpy(szTmp, Recent_Rom[i]);
        memmove(Recent_Rom[1], Recent_Rom[0], GMAX_PATH * i);
        strcpy(Recent_Rom[0], szTmp);
        memmove(Recent_Rom_Type + 1, Recent_Rom_Type, 4 * i);
        Recent_Rom_Type[0] = iType;
      }
    }
    else
    {
      memmove(Recent_Rom[1], Recent_Rom[0], GMAX_PATH * (GMAX_RECENT_ROM - 1));
      strcpy(Recent_Rom[0], pszRom);
      memmove(Recent_Rom_Type + 1, Recent_Rom_Type, 4 * (GMAX_RECENT_ROM - 1));
      Recent_Rom_Type[0] = iType;
    }
  }
}


void Update_Rom_Dir(char *Path)
{
  if(Use_Last_Open_Rom_Dir) Get_Dir_From_Path(Path, Rom_Dir);
}


void __stdcall Update_Rom_Name(LPCSTR Name)
{
  int i, leng;

  leng = strlen(Name) - 1;

  while ((leng >= 0) && (Name[leng] != '\\')) leng--;

  leng++; i = 0;

  while ((Name[leng]) && (Name[leng] != '.'))
    Rom_Name[i++] = Name[leng++];

  Rom_Name[i] = 0;
}


void Update_CD_Rom_Name(char *Name)
{
  int i, j;

  memcpy(Rom_Name, Name, 48);

  for(i = 0; i < 48; i++)
  {
    if((Rom_Name[i] >= '0') && (Rom_Name[i] <= '9')) continue;
    if(Rom_Name[i] == ' ') continue;
    if((Rom_Name[i] >= 'A') && (Rom_Name[i] <= 'Z')) continue;
    if((Rom_Name[i] >= 'a') && (Rom_Name[i] <= 'z')) continue;
    Rom_Name[i] = ' ';
  }

  for(i = 0; i < 48; i++)
  {
    if(Rom_Name[i] != ' ') i = 100;
  }

  if(i < 100) strcpy(Rom_Name, "no name");

  for(i = 47, j = 48; i >= 0; i--, j--)
  {
    if(Rom_Name[i] != ' ') i = -1;
  }

  Rom_Name[j + 1] = 0;
}


void De_Interleave(void)
{
  unsigned char buf[16384];
  unsigned char *Src;
  int i, j, Nb_Blocks, ptr;

  Src = &Rom_Data[0x200];

  Rom_Size -= 512;

  Nb_Blocks = Rom_Size / 16384;

  for(ptr = 0, i = 0; i < Nb_Blocks; i++, ptr += 16384)
  {
    memcpy(buf, Src + ptr, 16384);

    for(j = 0; j < 8192; j++)
    {
      Rom_Data[ptr + (j << 1) + 1] = buf[j];
      Rom_Data[ptr + (j << 1)]     = buf[j + 8192];
    }
  }
}


void __stdcall Fill_Infos_MD(void)
{
  int i;

  // Finally we do the IPS patch here, we can have the translated game name

  IPS_Patching();

  memset(&Rom_Info, 0, sizeof(GENESIS_ROM_INFO));

  for(i = 0; i < 16; i++)
    Rom_Info.Genesis.Console_Name[i] = Rom_Data[i + 256];

  for(i = 0; i < 16; i++)
    Rom_Info.Genesis.Copyright[i] = Rom_Data[i + 272];

  for(i = 0; i < 48; i++)
    Rom_Info.Genesis.Rom_Name[i] = Rom_Data[i + 288];

  for(i = 0; i < 48; i++)
    Rom_Info.Genesis.Rom_Name_W[i] = Rom_Data[i + 336];

  Rom_Info.Genesis.Type[0] = Rom_Data[384];
  Rom_Info.Genesis.Type[1] = Rom_Data[385];

  for(i = 0; i < 12; i++)
    Rom_Info.Genesis.Version[i] = Rom_Data[i + 386];

  Rom_Info.Genesis.Checksum = (Rom_Data[398] << 8) | Rom_Data[399];

  for(i = 0; i < 16; i++)
    Rom_Info.Genesis.IO_Support[i] = Rom_Data[i + 400];

  Rom_Info.Genesis.Rom_Start_Adress = Rom_Data[416] << 24;
  Rom_Info.Genesis.Rom_Start_Adress |= Rom_Data[417] << 16;
  Rom_Info.Genesis.Rom_Start_Adress |= Rom_Data[418] << 8;
  Rom_Info.Genesis.Rom_Start_Adress |= Rom_Data[419];

  Rom_Info.Genesis.Rom_End_Adress = Rom_Data[420] << 24;
  Rom_Info.Genesis.Rom_End_Adress |= Rom_Data[421] << 16;
  Rom_Info.Genesis.Rom_End_Adress |= Rom_Data[422] << 8;
  Rom_Info.Genesis.Rom_End_Adress |= Rom_Data[423];

  Rom_Info.Genesis.R_Size = Rom_Info.Genesis.Rom_End_Adress - Rom_Info.Genesis.Rom_Start_Adress + 1;

  for(i = 0; i < 12; i++)
    Rom_Info.Genesis.Ram_Infos[i] = Rom_Data[i + 424];

  Rom_Info.Genesis.Ram_Start_Adress = Rom_Data[436] << 24;
  Rom_Info.Genesis.Ram_Start_Adress |= Rom_Data[437] << 16;
  Rom_Info.Genesis.Ram_Start_Adress |= Rom_Data[438] << 8;
  Rom_Info.Genesis.Ram_Start_Adress |= Rom_Data[439];

  Rom_Info.Genesis.Ram_End_Adress = Rom_Data[440] << 24;
  Rom_Info.Genesis.Ram_End_Adress |= Rom_Data[441] << 16;
  Rom_Info.Genesis.Ram_End_Adress |= Rom_Data[442] << 8;
  Rom_Info.Genesis.Ram_End_Adress |= Rom_Data[443];

  for(i = 0; i < 12; i++)
    Rom_Info.Genesis.Modem_Infos[i] = Rom_Data[i + 444];

  for(i = 0; i < 40; i++)
    Rom_Info.Genesis.Description[i] = Rom_Data[i + 456];

  for(i = 0; i < 3; i++)
    Rom_Info.Genesis.Countries[i] = Rom_Data[i + 496];

  Rom_Info.Genesis.Console_Name[16] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Console_Name);
  Rom_Info.Genesis.Copyright[16] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Copyright);
  Rom_Info.Genesis.Rom_Name[48] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Rom_Name);
  Rom_Info.Genesis.Rom_Name_W[48] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Rom_Name_W);
  Rom_Info.Genesis.Type[2] = 0;
  Rom_Info.Genesis.Version[12] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Version);
  Rom_Info.Genesis.IO_Support[12] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.IO_Support);
  Rom_Info.Genesis.Ram_Infos[12] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Ram_Infos);
  Rom_Info.Genesis.Modem_Infos[12] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Modem_Infos);
  Rom_Info.Genesis.Description[40] = 0;
  Clear_String_Spaces(Rom_Info.Genesis.Description);
  Rom_Info.Genesis.Countries[3] = 0;
}


void __stdcall Fill_Infos_MS(void)
{
}


BOOL __stdcall Open_Rom(void)
{
  char         Name[GMAX_PATH];
  OPENFILENAME ofn;
  int          sys;
  BOOL         ret;
  GPLUGIN_ROM *pPlugin;

  SetCurrentDirectory(Gens_Dir);

  memset(Name, 0, GMAX_PATH);
  memset(&ofn, 0, sizeof(OPENFILENAME));

  if(!Open_Filter) Open_Filter = Create_Open_Filter();

  ofn.lStructSize     = sizeof(OPENFILENAME);
  ofn.hwndOwner       = ghWnd;
  ofn.hInstance       = ghInstance;
  ofn.lpstrFile       = Name;
  ofn.nMaxFile        = GMAX_PATH;
  ofn.lpstrFilter     = Open_Filter;
  ofn.nFilterIndex    = File_Type_Index;
  ofn.lpstrInitialDir = Rom_Dir;
  ofn.lpstrDefExt     = "smd";
  ofn.Flags           = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
  if(!GetOpenFileName(&ofn)) return FALSE;

  File_Type_Index = ofn.nFilterIndex;

  Free_Rom();

  // force plugin to load the rom
  if(File_Type_Index >= 7 && File_Type_Index < 7 + Rom_Plugins_Loaded)
    pPlugin = Rom_Plugins[File_Type_Index - 7];
  else // auto-choose the plugin to load the rom
    pPlugin = NULL;

  sys = Load_Rom(Name, pPlugin);
  if(!sys) return FALSE;

  if((File_Type_Index > 1) && (File_Type_Index < 7))
  {
    sys &= 1;
    sys |= (File_Type_Index - 1) << 1;
  }

  if(Game) Game->Type = sys;

  if(sys & 1) De_Interleave();
  Update_Rom_Name(Name);
  if     ((sys >> 1) < 3) Fill_Infos_MD();
  else if((sys >> 1) < 5) // SMS/GG rom
  {
    if((Rom_Size / 512) & 1) // skip image header
    { Rom_Size -= 512;
      memmove(Rom_Data, Rom_Data + 512, Rom_Size); }
    Fill_Infos_MS();
  }

  ret = FALSE;
  switch(sys >> 1)
  {
    default:
    case (GENESIS_ROM >> 1):    // Genesis rom
      if(Game) Genesis_Started = Init_Genesis(&Game->Genesis);
      ret = Genesis_Started;
      break;

    case (_32X_ROM >> 1):    // 32X rom
      if(Game) _32X_Started = Init_32X(&Game->Genesis);
      ret = _32X_Started;
      break;

    case (SMS_ROM >> 1):
    case (GG_ROM >> 1):
      if(Game) SMS_Started = SMS_Init(Game);
      ret = SMS_Started;
      break;

    case (SEGACD_IMAGE >> 1):    // Sega CD image
      SegaCD_Started = Init_SegaCD(Name);
      ret = SegaCD_Started;
      break;

    case (SEGACD_32X_IMAGE >> 1):    // Sega CD 32X image
      break;
  }

  // reading controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    sprintf(Str_Tmp, "%s%s%s", Config_Dir, Rom_Name, ".ini");
    Load_Control_Config(Str_Tmp);
  }

  if(ret)
  {
    Update_Recent_Rom(Name, sys);
    Update_Rom_Dir(Name);
    Build_Main_Menu();
    if(!Full_Screen && Switch_Full_Screen) Set_Render(ghWnd, 1, -1, Render_FS, TRUE);
  }

  return ret;
}


BOOL __stdcall Open_Recent_Rom(LPSTR Name)
{
  int    sys;
  BOOL   ret;

  SetCurrentDirectory(Gens_Dir);

  if(GetFileAttributes(Name) & FILE_ATTRIBUTE_DIRECTORY) return 0;

  Free_Rom();

  sys = Load_Rom(Name, NULL);
  if(!sys) return FALSE;

  if(sys & 1) De_Interleave();
  Update_Rom_Name(Name);
  if     ((sys >> 1) < 3) Fill_Infos_MD();
  else if((sys >> 1) < 5) // SMS/GG rom
  {
    if((Rom_Size / 512) & 1) // skip image header
    { Rom_Size -= 512;
      memmove(Rom_Data, Rom_Data + 512, Rom_Size); }
    Fill_Infos_MS();
  }

  if(Game) Game->Type = sys;

  ret = FALSE;
  switch(sys >> 1)
  {
    default:
    case (GENESIS_ROM >> 1):    // Genesis rom
      if(Game) Genesis_Started = Init_Genesis(&Game->Genesis);
      ret = Genesis_Started;
      break;

    case (_32X_ROM >> 1):    // 32X rom
      if(Game) _32X_Started = Init_32X(&Game->Genesis);
      ret = _32X_Started;
      break;

    case (SMS_ROM >> 1):
    case (GG_ROM >> 1):
      if(Game) SMS_Started = SMS_Init(Game);
      ret = SMS_Started;
      break;

    case (SEGACD_IMAGE >> 1):    // Sega CD image
      SegaCD_Started = Init_SegaCD(Name);
      ret = SegaCD_Started;
      break;

    case (SEGACD_32X_IMAGE >> 1):    // Sega CD 32X image
      break;
  }

  // reading controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    sprintf(Str_Tmp, "%s%s%s", Config_Dir, Rom_Name, ".ini");
    Load_Control_Config(Str_Tmp);
  }

  if(ret)
  {
    Update_Recent_Rom(Name, sys);
    Update_Rom_Dir(Name);
    Build_Main_Menu();
    if(!Full_Screen && Switch_Full_Screen) Set_Render(ghWnd, 1, -1, Render_FS, TRUE);
  }

  return ret;
}


// Rom is already in buffer, we just need to fill rom structure
// and do init stuff...
int __stdcall Load_Rom_CC(LPCSTR Name, int Size)
{
  Update_Rom_Name(Name);
  Rom_Size = Size;
  Fill_Infos_MD();

  Game = &Rom_Info;

  Genesis_Started = Init_Genesis(&Game->Genesis);
  Build_Main_Menu();

  return Genesis_Started;
}


BOOL __stdcall Load_Bios(LPSTR Name)
{
  int sys;

  Free_Rom();
  SetCurrentDirectory(Gens_Dir);
  sys = Load_Rom(Name, NULL);
  if(sys & 1) De_Interleave();

  return !!sys;
}


LPVOID __stdcall Load_SMS_Bios(void)
{
  HRSRC   hRes;
  HGLOBAL hData;

  hRes = FindResource(ghInstance, MAKEINTRESOURCE(ID_SMSBIOS), RT_RCDATA);
  if(!hRes) return NULL;
  hData = LoadResource(ghInstance, hRes);
  if(!hData) return NULL;

  return LockResource(hData);
}


// load a rom from file, returns the rom type or 0 if error
int __stdcall Load_Rom(LPSTR lpFile, GPLUGIN_ROM *pPlugin)
{
  HANDLE        hFile;
  DWORD         dwRead;
  char          pName[GMAX_PATH], pPath[GMAX_PATH];
  LPSTR         pExt;
  HUNZFILE      zFile;
  LPSTR         pBuf = (LPSTR)Rom_Data;
  UNZ_FILE_INFO zInfo;
  GZFILE        gFile;
  int           iSize = 0;
  int           i, iExtType = 0;
  
  SetCurrentDirectory(Gens_Dir);
  dwRead = GetFullPathName(lpFile, GMAX_PATH, pPath, &pExt);
  if(!dwRead || dwRead >= GMAX_PATH) strcpy(pPath, lpFile);

  memset(pBuf, 0, 6 * 1024 * 1024);
  // check if there are any valid plugins to open the file
  if(!pPlugin)
  {
    for(i = 0; i < Rom_Plugins_Loaded; i++)
    {
      if(Rom_Plugins[i]->Check(pPath)) { pPlugin = Rom_Plugins[i]; break; }
    }
  }

  if(pPlugin) // must use a plugin to open the rom
  {
    SetCurrentDirectory(Plugin_Dir);

    hFile = pPlugin->Open(pPath);
    if(!hFile) return 0;
    i = 0;
    while(pPlugin->EnumFiles(hFile, i, pName, GMAX_PATH))
    {
      pExt = Get_File_Ext(pName);
      if(pExt)
      {
        if(!strcmpi(".SMD", pExt) || !strcmpi(".BIN", pExt)  || !strcmpi(".GEN", pExt))
          iExtType = GENESIS_ROM;
        else if(!strcmpi(".32X", pExt)) iExtType = _32X_ROM;
        else if(!strcmpi(".SMS", pExt)) iExtType = SMS_ROM;
        else if(!strcmpi(".GG", pExt))  iExtType = GG_ROM;
        if(iExtType)
        { iSize = pPlugin->LoadFile(hFile, i, pName, pBuf, 6 * 1024 * 1024);
          break; }
      }
      i++;
    }
    pPlugin->Close(hFile);
  }
  else
  {
    pExt = Get_File_Ext(pPath);
    if(pExt && (!strcmpi(pExt, ".ZIP") || !strcmpi(pExt, ".ZSG"))) // zipfile
    {
      zFile = unzOpen(pPath);
      if(!zFile) return 0;
      do if(unzGoToFirstFile(zFile) == UNZ_OK)
      {
        unzGetCurrentFileInfo(zFile, &zInfo, pName, GMAX_PATH, NULL, 0, NULL, 0);
        pExt = Get_File_Ext(pName);
        if(pExt)
        {
          if(!strcmpi(".SMD", pExt) || !strcmpi(".BIN", pExt)  || !strcmpi(".GEN", pExt))
            iExtType = GENESIS_ROM;
          else if(!strcmpi(".32X", pExt)) iExtType = _32X_ROM;
          else if(!strcmpi(".SMS", pExt)) iExtType = SMS_ROM;
          else if(!strcmpi(".GG", pExt))  iExtType = GG_ROM;
          if(iExtType)
          {
            if(zInfo.dwUncompSize <= (6 * 1024 * 1024))
            {
              if(unzOpenCurrentFile(zFile) == UNZ_OK)
              {
                iSize = unzReadCurrentFile(zFile, pBuf, zInfo.dwUncompSize);
                unzCloseCurrentFile(zFile);
              }
            }
            break;
          }
        }
      } while(unzGoToNextFile(zFile) == UNZ_OK);

      unzClose(zFile);
    }
    else // not a zipfile
    {
      if(!strcmpi(".SMD", pExt) || !strcmpi(".BIN", pExt)  || !strcmpi(".GEN", pExt))
        iExtType = GENESIS_ROM;
      else if(!strcmpi(".32X", pExt)) iExtType = _32X_ROM;
      else if(!strcmpi(".SMS", pExt)) iExtType = SMS_ROM;
      else if(!strcmpi(".GG", pExt))  iExtType = GG_ROM;

      hFile = CreateFile(pPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
      if(hFile == INVALID_HANDLE_VALUE) return 0;
      if(!ReadFile(hFile, pBuf, 1024, &dwRead, NULL) || dwRead != 1024)
      { CloseHandle(hFile); return 0; }

      if(pBuf[0] == 0x1F && pBuf[1] == 0x8B) // gzfile
      {
        CloseHandle(hFile);
        gFile = gzopen(pPath, "rb");
        if(!gFile) return 0;
        iSize = gzread(gFile, pBuf, 6 * 1024 * 1024);
        if(!gzeof(gFile)) iSize = 0;
        gzclose(gFile);
      }
      else if(!strnicmp("SEGADISCSYSTEM", pBuf, 14))
      { CloseHandle(hFile); return SEGACD_IMAGE; }     // Sega CD (ISO)
      else if(!strnicmp("SEGADISCSYSTEM", pBuf + 0x10, 14))
      { CloseHandle(hFile); return SEGACD_IMAGE + 1; } // Sega CD (BIN)
      else
      {
        iSize = GetFileSize(hFile, NULL);
        if(iSize > (6 * 1024 * 1024)) { CloseHandle(hFile); return 0; }
        iSize -= 1024;
        if(!ReadFile(hFile, pBuf + 1024, iSize, &dwRead, NULL) || dwRead != iSize) iSize = 0;
        else iSize += 1024;
        CloseHandle(hFile);
      }
    }
  }

  if(iSize <= 0) return 0; // error on read the file

  Game = &Rom_Info;
  Rom_Size = iSize;

  if(!memcmp(pBuf + 0x7FF0, "TMR SEGA", 8))
  {
    if(iExtType == GG_ROM) return GG_ROM;
    else                   return SMS_ROM;
  }

  i = 0;

  if(strnicmp("SEGA", pBuf + 0x100, 4))
  {
    // Maybe interleaved
    if(!strnicmp("EA", pBuf + (0x200 + (0x100 / 2)), 2))               i = 1;
    if(pBuf[0x08] == 0xAA && pBuf[0x09] == 0xBB && pBuf[0x0A] == 0x06) i = 1;
  }

  if(i)    // interleaved
  {
    if(iExtType == _32X_ROM && pBuf[0x200 / 2] == 0x4E)  return _32X_ROM + 1;
    if(!strnicmp("3X", pBuf + (0x200 + (0x105 / 2)), 3)) return _32X_ROM + 1;
  }
  else
  {
    if(iExtType == _32X_ROM && pBuf[0x200] == 0x4E) return _32X_ROM;
    if(!strnicmp("32X", pBuf + 0x105, 3))           return _32X_ROM;
  }

  return GENESIS_ROM + i;
}


WORD Calculate_Checksum(void)
{
  unsigned short checksum = 0;
  unsigned int i;

  if(!Game) return(0);

  for (i = 512; i < Rom_Size; i += 2)
  {
    checksum += (unsigned short)(Rom_Data[i + 0]);
    checksum += (unsigned short)(Rom_Data[i + 1] << 8);
  }

  return checksum;
}


void Fix_Checksum(void)
{
  unsigned short checks;

  if(!Game) return;

  checks = Calculate_Checksum();

  if(Rom_Size)
  {
    Rom_Data[0x18E] = checks & 0xFF;
    Rom_Data[0x18F] = checks >> 8;
    _32X_Rom[0x18E] = checks >> 8;
    _32X_Rom[0x18F] = checks & 0xFF;
    if(Game) Game->Genesis.Checksum = (Rom_Data[0x18F] << 8) | Rom_Data[0x18E];
  }
}


unsigned int Calculate_CRC32(void)
{
  unsigned int crc = 0;

  Byte_Swap(Rom_Data, Rom_Size);
  crc = crc32(0, Rom_Data, Rom_Size);
  Byte_Swap(Rom_Data, Rom_Size);

  return crc;
}


int IPS_Patching(void)
{
  FILE *IPS_File;
  char Name[1024];
  unsigned char buf[16];
  unsigned int adr, len, i;

  SetCurrentDirectory(Gens_Dir);

  strcpy(Name, IPS_Dir);
  strcat(Name, Rom_Name);
  strcat(Name, ".ips");

  IPS_File = fopen(Name, "rb");

  if(IPS_File == NULL) return 1;

  fseek(IPS_File, 0, SEEK_SET);

  fread(buf, 1, 5, IPS_File);
  buf[5] = 0;

  if(strcmpi((char *) buf, "patch"))
  {
    fclose(IPS_File);
    return 2;
  }

  fread(buf, 1, 3, IPS_File);
  buf[3] = 0;

  while (strcmpi((char *) buf, "eof"))
  {
    adr = (unsigned int) buf[2];
    adr += (unsigned int) (buf[1] << 8);
    adr += (unsigned int) (buf[0] << 16);

    if(fread(buf, 1, 2, IPS_File) == 0)
    {
      fclose(IPS_File);
      return 3;
    }

    len = (unsigned int) buf[1];
    len += (unsigned int) (buf[0] << 8);

    for(i = 0; i < len; i++)
    {
      if(fread(buf, 1, 1, IPS_File) == 0)
      {
        fclose(IPS_File);
        return 3;
      }

      if(adr < Rom_Size) Rom_Data[adr++] = buf[0];
    }

    if(fread(buf, 1, 3, IPS_File) == 0)
    {
      fclose(IPS_File);
      return 3;
    }

    buf[3] = 0;
  }

  fclose(IPS_File);

  return 0;
}


void __stdcall Free_Rom(void)
{
  if     (Movie_Status & MOVIE_PLAYING)   StopPlayMovie();
  else if(Movie_Status & MOVIE_RECORDING) StopRecordMovie();

  if(!Game) return;

#ifdef CC_SUPPORT
  CC_Close();
#endif

  if(SegaCD_Started) Save_BRAM();
  if(SMS_Started)    SMS_End_Emulation();
  else               Save_SRAM();
  Save_Patch_File();
  if(WAV_Dumping) Stop_WAV_Dump();
  if(GYM_Dumping) Stop_GYM_Dump();
  if(GSR_Dumping) Stop_GSR_Dump();
  if(VGM_File)    Stop_VGM_Dump();
  if(SegaCD_Started) Stop_CD();
  Net_Play = 0;
  Genesis_Started = 0;
  _32X_Started    = 0;
  SegaCD_Started  = 0;
  SMS_Started     = 0;
  Game = NULL;

  Reset_VDP();
  if(Intro_Style == 3) Init_Genesis_Bios();

  // saving controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    FILE *fcfg;
    strcpy(Str_Tmp, Config_Dir);
    strcat(Str_Tmp, Rom_Name);
    strcat(Str_Tmp, ".ini");
    fcfg = fopen(Str_Tmp, "wt");
    if(fcfg)
    {
      Save_Control_Config(fcfg);
      fclose(fcfg);
    }
    // reading the default controller configuration
    sprintf(Str_Tmp, "%s%s", Gens_Dir, "gens.ini");
    Load_Control_Config(Str_Tmp);
  }

  Set_Title(NULL, NULL);
  memset(Rom_Name, 0, sizeof(Rom_Name));
}



const char def_open_filter[] =
  "All suported files\0*.bin;*.smd;*.gen;*.32x;*.sms;*.gg;*.iso;*.raw;*.zip;*.zsg;\0" \
  "Genesis roms (smd, bin, gen, zip, zsg)\0*.smd;*.bin;*.gen;*.zip;*.zsg;\0"          \
  "32X roms (32x, zip)\0*.32x;*.zip;\0"                                               \
  "Master System roms (sms, zip)\0*.sms;*.zip;\0"                                     \
  "Game Gear roms (gg, zip)\0*.gg;*.zip;\0"                                           \
  "Sega CD images (iso, bin, raw)\0*.iso;*.bin;*.raw\0"                               \
  "All files (*.*)\0*.*\0";

const char null_filter[]  = " \0 ";
const char valid_filter[] = "All suported files\0*.bin;*.smd;*.gen;*.32x;*.sms;*.gg;*.iso;*.raw;*.zip;*.zsg;";
const char type_filter[]  = 
  "Genesis roms (smd, bin, gen, zip, zsg)\0*.smd;*.bin;*.gen;*.zip;*.zsg;\0" \
  "32X roms (32x, zip)\0*.32x;*.zip;\0"                                      \
  "Master System roms (sms, zip)\0*.sms;*.zip;\0"                            \
  "Game Gear roms (gg, zip)\0*.gg;*.zip;\0"                                  \
  "Sega CD images (iso, bin, raw)\0*.iso;*.bin;*.raw";
const char all_filter[]   = "All files (*.*)\0*.*\0";


LPCSTR __stdcall Create_Open_Filter(void)
{
  LPSTR s, f, ret;
  int   i, size, b, l[2];

  // calculating total size of the string
  for(i = size = 0; i < Rom_Plugins_Loaded; i++)
  {
    f = (LPSTR)Rom_Plugins[i]->lpOpenFilter;
    if(!f) f = (LPSTR)null_filter;
    l[0] = l[1] = b = 0;
    while(b < 2) { if(!f[l[b]++]) f += l[b++]; }
    size += l[0] + (l[1] << 1);
  }
  size += sizeof(valid_filter) + sizeof(type_filter) + sizeof(all_filter) - 1;

  // alloc memory
  s = (LPSTR)_alloc(size);
  if(!s) return def_open_filter;
  ret = s;

  // creating the string
  memcpy(s, valid_filter, sizeof(valid_filter) - 1);
  s += sizeof(valid_filter) - 1;
  for(i = 0; i < Rom_Plugins_Loaded; i++)
  {
    f = (LPSTR)Rom_Plugins[i]->lpOpenFilter;
    if(!f) f = (LPSTR)null_filter;
    while((*f++));
    while((*f)) *s++ = *f++;
    *s++ = ';';
  }
  *(s - 1) = 0;

  memcpy(s, type_filter, sizeof(type_filter));
  s += sizeof(type_filter);

  for(i = 0; i < Rom_Plugins_Loaded; i++)
  {
    f = (LPSTR)Rom_Plugins[i]->lpOpenFilter;
    if(!f) f = (LPSTR)null_filter;
    b = 0;
    while(b < 2)
    { *s++ = *f; if(!(*f++)) b++; }
  }

  memcpy(s, all_filter, sizeof(all_filter));

  return (LPCSTR)ret;
}
