
#include "gmain.h"
#include "gplugins.h"
#include "config.h"
#include "gutil.h"
#include <string.h>

// used to alloc memory
#define _alloc(size)          malloc(size)
#define _realloc(ptr, size) ( ptr ? realloc(ptr, size) : malloc(size) )
#define _free(ptr)            if(ptr) { free(ptr); ptr = NULL; }


int                Plugins_Loaded          = 0;
GPLUGIN          **Plugins                 = NULL;
int                Rom_Plugins_Loaded      = 0;
GPLUGIN_ROM      **Rom_Plugins             = NULL;
int                Shot_Types_Loaded       = 0;
int                Shot_Plugins_Loaded     = 0;
GPLUGIN_SHOT     **Shot_Plugins            = NULL;
int                VideoLog_Plugins_Loaded = 0;
GPLUGIN_VIDEOLOG **VideoLog_Plugins        = NULL;

// load all plugins on the plugins folder
BOOL __stdcall Init_Plugins(void)
{
  HANDLE          hFind;
  WIN32_FIND_DATA wfd;
  GPLUGIN        *pData;
  LPVOID          ptr;
  char            pCfgDir[MAX_PATH + 4];
  LPSTR           pTmp;
  DWORD           l;

  SetCurrentDirectory(Gens_Dir);
  l = GetFullPathName(Config_Dir, MAX_PATH, pCfgDir, &pTmp);
  if(!l || l >= MAX_PATH)
  { l = strlen(Gens_Dir); memcpy(pCfgDir, Gens_Dir, l + 1); }

  if(pCfgDir[--l] != '\\' && pCfgDir[l] != '/') pCfgDir[++l] = '\\';

  if(!SetCurrentDirectory(Plugin_Dir)) return TRUE;

  hFind = FindFirstFile("*.DLL", &wfd);
  if(hFind == INVALID_HANDLE_VALUE) return TRUE;

  do
  {
    pData = Load_Plugin(wfd.cFileName, pCfgDir);
    if(pData)
    {
      if(pData->dwType == GPLUGIN_TYPE_ROM)
      {
        ptr = _realloc(Rom_Plugins, (Rom_Plugins_Loaded + 1) * 4);
        if(ptr)
        {
          Rom_Plugins = (GPLUGIN_ROM **)ptr;
          Rom_Plugins[Rom_Plugins_Loaded++] = (GPLUGIN_ROM *)pData;
        }
      }
      else if(pData->dwType == GPLUGIN_TYPE_SHOT)
      {
        ptr = _realloc(Shot_Plugins, (Shot_Plugins_Loaded + 1) * 4);
        if(ptr)
        {
          Shot_Plugins = (GPLUGIN_SHOT **)ptr;
          Shot_Plugins[Shot_Plugins_Loaded++] = (GPLUGIN_SHOT *)pData;
          Shot_Types_Loaded += ((GPLUGIN_SHOT *)pData)->nShot;
        }
      }
      else if(pData->dwType == GPLUGIN_TYPE_VIDEOLOG)
      {
        ptr = _realloc(VideoLog_Plugins, (VideoLog_Plugins_Loaded + 1) * 4);
        if(ptr)
        {
          VideoLog_Plugins = (GPLUGIN_VIDEOLOG **)ptr;
          VideoLog_Plugins[VideoLog_Plugins_Loaded++] = (GPLUGIN_VIDEOLOG *)pData;
        }
      }
      else // invalid kind of plugin
        Free_Plugin(pData);

      ptr = _realloc(Plugins, (Plugins_Loaded + 1) * 4);
      if(ptr) 
      {
        Plugins = (GPLUGIN **)ptr;
        Plugins[Plugins_Loaded++] = (GPLUGIN *)pData;
      }
    }
  } while(FindNextFile(hFind, &wfd));

  return TRUE;
}


// free all loaded plugins
void __stdcall Quit_Plugins(void)
{
  int i;

  SetCurrentDirectory(Plugin_Dir);

  for(i = 0; i < Plugins_Loaded; i++) Free_Plugin(Plugins[i]);
  
  Plugins_Loaded = Rom_Plugins_Loaded = Shot_Plugins_Loaded = Shot_Types_Loaded = 
  VideoLog_Plugins_Loaded = 0;

  _free(VideoLog_Plugins);
  _free(Shot_Plugins);
  _free(Rom_Plugins);
  _free(Plugins);
}



typedef GPLUGIN *(__stdcall *GETPLUGININTERFACEPROC)(void);

// Loads a plugin. Returns the plugin data on success. On error returns NULL.
GPLUGIN *__stdcall Load_Plugin(LPCSTR lpFileName, LPCSTR lpCfgDir)
{
  HINSTANCE              hInstDLL;
  GPLUGIN               *pPlugin;
  GETPLUGININTERFACEPROC GetPluginData;
  char                   pName[GMAX_PATH], pPath[GMAX_PATH];
  LPSTR                  pConfig;
  LPSTR                  pExt;
  int                    len;

  hInstDLL = LoadLibrary(lpFileName);
  if(!hInstDLL) return NULL;

  GetPluginData = (GETPLUGININTERFACEPROC)GetProcAddress(hInstDLL, "_GensPlugin_GetInterface@0");
  if(!GetPluginData) { FreeLibrary(hInstDLL); return NULL; }
  pPlugin = GetPluginData();
  if(!pPlugin)   { FreeLibrary(hInstDLL); return NULL; }
  pPlugin->hInstance = hInstDLL;
  pPlugin->hWnd      = ghWnd;

  if(!GetFileTitle(lpFileName, pName, GMAX_PATH))
  {
    pExt = Get_File_Ext(pName);
    if(pExt) *pExt = 0;
    len = sprintf(pPath, "%s%s%s", lpCfgDir, pName, ".ini") + 1;
  }
  else strcpy(pPath, "plugin.ini");

  pConfig = (LPSTR)_alloc(len);
  if(pConfig) memcpy(pConfig, pPath, len);
  pPlugin->lpConfigFileName = (LPCSTR)pConfig;

  if(pPlugin->Init)
  {
    if(!pPlugin->Init())
    {
      FreeLibrary(hInstDLL);
      _free(pConfig);
    }
  }

  return pPlugin;
}

void __stdcall Free_Plugin(GPLUGIN *pPlugin)
{
  LPVOID ptr;

  ptr = (LPVOID)pPlugin->lpConfigFileName;
  if(pPlugin->Quit) pPlugin->Quit();
  FreeLibrary(pPlugin->hInstance);
  _free(ptr);
}
