
#ifndef _PLUGIN_H
#define _PLUGIN_H

#include <windows.h>

// info common to all kinds of plugins, all plugins struct begins with this
typedef struct _GPLUGIN {
  DWORD      dwType;                       // type of the plugin
  HINSTANCE  hInstance;                    // handle of the plugin instance         (filled by gens)
  HWND       hWnd;                         // handle of gens main window            (filled by gens)
  LPCSTR     lpConfigFileName;             // path to configuration file to be used (filled by gens)
  LPCSTR     lpDisplayName;                // plugin name to be displayed on menus
  BOOL (__stdcall *Init)(void);            // optional initialization function, must return TRUE if ok
  void (__stdcall *Quit)(void);            // optional end function
  void (__stdcall *About)(void);           // shows plugin's about dialog         (optional)
  void (__stdcall *Config)(void);          // shows plugin's configuration dialog (optional)
} GPLUGIN;

#ifdef __cplusplus

// rom compression plugins
typedef struct _GPLUGIN_ROM : _GPLUGIN {
  LPCSTR     lpOpenFilter;                 // open dialog filter
  BOOL   (__stdcall *Check)(LPCSTR);       // check if it's a valid file
  HANDLE (__stdcall *Open)(LPCSTR);        // open the file, must return INVALID_HANDLE_VALUE if error
  void   (__stdcall *Close)(HANDLE);       // close the file, HANDLE is the value returned by Open
  BOOL   (__stdcall *EnumFiles)(HANDLE, UINT, LPSTR, UINT);
  int    (__stdcall *LoadFile)(HANDLE, UINT, LPCSTR, LPVOID, UINT);
} GPLUGIN_ROM;


typedef struct _GSHOT_TYPE {
  LPCSTR     lpDisplayName;                // shot name to be displayed on menu
  LPCSTR     lpExt;                        // extension to use
  BOOL   (__stdcall *SaveImage)(LPCSTR, BITMAPINFO *, LPBYTE);
} GSHOT_TYPE;

// screen shot plugins
typedef struct _GPLUGIN_SHOT : _GPLUGIN {
  UINT        nShot;                       // total number of shots avaible on plugin
  GSHOT_TYPE *Shots;
} GPLUGIN_SHOT;


// video record plugins
typedef struct _GPLUGIN_VIDEOLOG : _GPLUGIN {
  LPCSTR      lpSaveFilter;                // save dialog filter
  LPCSTR      lpDefExt;                    // default extension to use on files
  HANDLE (__stdcall *Create)(LPCSTR, int, int, int);
  void   (__stdcall *Close)(HANDLE);
  BOOL   (__stdcall *SaveFrame)(HANDLE, LPWORD, int, int, int);
} GPLUGIN_VIDEOLOG;

#else

typedef struct GPLUGIN GPLUGIN_ROM;
typedef struct GPLUGIN GPLUGIN_SHOT;
typedef struct GPLUGIN GPLUGIN_VIDEOLOG;

#endif

// valid plugin types
#define GPLUGIN_TYPE_ROM      0x01             // plugin for rom compress
#define GPLUGIN_TYPE_SHOT     0x02             // plugin for screen shot
#define GPLUGIN_TYPE_VIDEOLOG 0x03             // plugin for video record

// plugins loaded
extern int                Plugins_Loaded;
extern GPLUGIN          **Plugins;
extern int                Rom_Plugins_Loaded;
extern GPLUGIN_ROM      **Rom_Plugins;
extern int                Shot_Types_Loaded;
extern int                Shot_Plugins_Loaded;
extern GPLUGIN_SHOT     **Shot_Plugins;
extern int                VideoLog_Plugins_Loaded;
extern GPLUGIN_VIDEOLOG **VideoLog_Plugins;

// load all plugins on the plugins folder
BOOL     __stdcall Init_Plugins(void);
// free all loaded plugins
void     __stdcall Quit_Plugins(void);
// loads a plugin, returns NULL if fails
GPLUGIN *__stdcall Load_Plugin(LPCSTR, LPCSTR);
// release a plugin
void     __stdcall Free_Plugin(GPLUGIN *);

#endif
