/* gsr.cpp - projeto gens+. Rodrigo Cardoso.
   Loga arquivos GSR. O formato VGM fica muito grande para genesis, estou tentando criar
   um formato que fique menor. */

/* lista de comandos:
   EX b1    - YM2612 port 1, escreve valor b1 para o reg 0x2X
   F0       - YM2612 e PSG update (fim do frame)
   F7 b1 b2 - YM2612 port 1, escreve valor b2 para o reg b1 e muda cur_a_adr para b1
   F8 b1 b2 - YM2612 port 2, escreve valor b2 para o reg b1 e muda cur_b_adr para b1
   F9 bb    - YM2612 port 1, escreve valor bb para cur_a_adr
   FA bb    - YM2612 port 2, escreve valor bb para cur_b_adr
   FB bb    - PSG, escreve valor bb
   FC w1 w2 - abre o sample de ndice w1 na tabela e inicia e atribui w2 para linha
              atual do vdp
   FF       - EOF */

#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include "zlib.h"
#include "gmain.h"
#include "grom.h"
#include "config.h"
#include "gsr.h"

//#define GSR_DAC_LOG

#define _alloc(size) malloc(size)
#define _realloc(ptr, size) ( ptr ? realloc(ptr, size) : malloc(size) )
#define _free(ptr)  { if(ptr) free(ptr); ptr = NULL; }

typedef struct _SAMPLE_BUF {
  LPBYTE          ptr;
  int             buf_size, size;
  __inline void   init(void);
  __inline void   quit(void);
  BOOL __fastcall expand(int);
  __inline BOOL   put_byte(int);
  __inline void   reset(void);
} SAMPLE_BUF;

__inline void _SAMPLE_BUF :: init(void)
{
  ptr  = NULL;
  size = buf_size = 0;
}

__inline void _SAMPLE_BUF :: quit(void)
{
  _free(ptr);
  init();
}

BOOL __fastcall _SAMPLE_BUF :: expand(int s)
{
  LPVOID p;
  p = _realloc(ptr, s + buf_size);
  if(!p) return FALSE;
  ptr = (LPBYTE)p;
  buf_size += s;
  return TRUE;
}

__inline BOOL _SAMPLE_BUF :: put_byte(int c)
{
  if(buf_size <= size) { if(!expand(1024)) return FALSE; }
  ptr[size++] = (BYTE)c;
  return TRUE;
}

 __inline void _SAMPLE_BUF :: reset(void)
{
//  if(ptr) memset(ptr, 0, buf_size);
  size = 0;
}

typedef struct _SAMPLE_DATA {
  int    size;
  LPBYTE ptr;
} SAMPLE_DATA;

typedef struct _SAMPLE_TABLE {
  SAMPLE_DATA    *data;
  int             nptr;
  int             ndata;
  __inline void   init(void);
  __inline void   quit(void);
  BOOL __fastcall add_data(LPBYTE, int);
  void __fastcall delete_data(int);
  void __fastcall swap_data(int, int);
} SAMPLE_TABLE;

__inline void _SAMPLE_TABLE :: init(void)
{
  nptr = ndata = 0;
  data = NULL;
}

__inline void _SAMPLE_TABLE :: quit(void)
{
  for(int i = 0; i < ndata; i++) _free(data[i].ptr);
  _free(data);
  init();
}

BOOL __fastcall _SAMPLE_TABLE :: add_data(LPBYTE d, int s)
{
  LPVOID p;
  if(nptr <= ndata)
  { p = _realloc(data, (nptr + 256) * sizeof(SAMPLE_DATA));
    if(!p) return FALSE;
    data = (SAMPLE_DATA *)p;
    nptr += 256; }
  p = _alloc(s);
  if(!p) return FALSE;
  memcpy(p, d, s);
  data[ndata].size = s;
  data[ndata].ptr  = (LPBYTE)p;
  ndata++;
  return TRUE;
}

void __fastcall _SAMPLE_TABLE :: delete_data(int n)
{
  if(n >= ndata) return;
  _free(data[n].ptr);
  data[n].size = 0;
//  if(n < --ndata) memcpy(data + n, data + n + 1, (ndata - n) * sizeof(SAMPLE_DATA));
}

void __fastcall _SAMPLE_TABLE :: swap_data(int d1, int d2)
{
  SAMPLE_DATA buf;
  if(d1 >= ndata || d2 >= ndata || d1 < 0 || d2 < 0) return;
  buf      = data[d1];
  data[d1] = data[d2];
  data[d2] = buf;
}


extern "C" int  CPU_Mode;
extern "C" BYTE State_Buffer[];
extern "C" int  VDP_Current_Line;
int             GSR_Dumping = 0;
FILE           *fdata;
int             gsr_file;
UINT            n_frames;
CHIPSINFO       ci;
int             dac_playing, dac_changed, last_dac;
int             cur_a_adr, cur_b_adr;
SAMPLE_BUF      sample_buf, true_sample_buf;
SAMPLE_TABLE    samples_table, true_samples_table;

#ifdef GSR_DAC_LOG
FILE *fDACLog = NULL;
#endif

// cria o arquivo e inicia o cabealho
void __stdcall Start_GSR_Dump(void)
{
  LPSTR      sufix;
  const char ext[] = "gsr";
  int        i, j, d;

  if(GSR_Dumping) Stop_GSR_Dump();

  if(!(*Rom_Name)) return;

  i     = sprintf(Str_Tmp, "%s%s", Dump_Dir, Rom_Name);
  sufix = Str_Tmp + i;
  d     = -1;
  do
  {
    if (d++ > 99999) return;
    sufix[0] = '_';
    i = 1;
    j = d / 10000;
    if (j) sufix[i++] = '0' + j;
    j = (d / 1000) % 10;
    if (j) sufix[i++] = '0' + j;
    j = (d / 100) % 10;
    sufix[i++] = '0' + j;
    j = (d / 10) % 10;
    sufix[i++] = '0' + j;
    j = d % 10;
    sufix[i++] = '0' + j;
    sufix[i++] = '.';
    sufix[i++] = ext[0];
    sufix[i++] = ext[1];
    sufix[i++] = ext[2];
    sufix[i]   = 0;
    // tentando criar o arquivo de sada
    gsr_file = open(Str_Tmp, _O_CREAT | _O_EXCL | _O_WRONLY | _O_BINARY | _O_SEQUENTIAL);
  } while(gsr_file == -1);

  // criando arquivos temporrios
  i = GetTempPath(sizeof(Str_Tmp) / 2, Str_Tmp + (sizeof(Str_Tmp) / 2));
  if(i == 0 || i > (sizeof(Str_Tmp) / 2))
  { strcpy(Str_Tmp + (sizeof(Str_Tmp) / 2), "."); }
  GetTempFileName(Str_Tmp + (sizeof(Str_Tmp) / 2), "tmp", 0, Str_Tmp);
  fdata = fopen(Str_Tmp, "w+bD");
  if(!fdata) { close(gsr_file); return; }

  // salvando o estado inicial dos chips de som
  for(i = 0; i < 8; i++) ci.PSG_Reg[i] = PSG.Register[i];
  for(i = 0; i < 0x100; i++)
  {
    ci.YM2612_Reg[0x000 + i] = YM2612.REG[0][i];
    ci.YM2612_Reg[0x100 + i] = YM2612.REG[1][i];
  }

  cur_a_adr   = cur_b_adr   = 0;
  dac_changed = 0;
  dac_playing = 0;
  n_frames    = 0;

  sample_buf.init();
  true_sample_buf.init();
  samples_table.init();
  true_samples_table.init();

#ifdef GSR_DAC_LOG
  fDACLog = fopen("gsr_dac.log", "wt");
  if(fDACLog) fprintf(fDACLog, "VDP - DAC\n\n");
#endif

  GSR_Dumping = 1;
}


#define gsrputc(c) putc(c, fdata)
#define gsrputw(w) gsrputc(((w >> 8) & 0xFF)); gsrputc((w & 0xFF))

#define add_dac_data(data) if(sample_buf.size < 1) {                   \
  gsrputc(0xFC);                                                       \
  gsrputw(samples_table.ndata);                                        \
  gsrputw(VDP_Current_Line); }                                         \
  sample_buf.put_byte(data)

#define flush_dac_data() if(sample_buf.size > 0) {                        \
  samples_table.add_data(sample_buf.ptr, sample_buf.size);                \
  sample_buf.reset();                                                     \
  true_samples_table.add_data(true_sample_buf.ptr, true_sample_buf.size); \
  true_sample_buf.reset(); }                                              \


// adiciona comando
void __stdcall Update_GSR_Dump(int cmd, int d1, int d2)
{
  switch(cmd)
  {
    case GSRDUMP_FRAME_END:
      if(!dac_changed) { flush_dac_data(); dac_playing = 0; }
      gsrputc(0xF0);
      dac_changed = 0;
      n_frames++;
      break;
    case GSRDUMP_CHANGE_VDP_LINE: // incrementa a linha do vdp
      if(dac_playing) true_sample_buf.put_byte(last_dac);
      break;
    case GSRDUMP_YM2612_PORT1:
      if(d1 == 0x2A) // dac
      { 
        if(YM2612.DAC)
        {
          dac_playing = dac_changed = 1; add_dac_data(d2);
          last_dac = d2;
#ifdef GSR_DAC_LOG
          if(fDACLog) fprintf(fDACLog, "%.3i - 0x%.2X\n", VDP_Current_Line, d2);
#endif
        }
      }
      else if((d1 & 0xF0) == 0x20)
      { gsrputc((d1 & 0x0F) | 0xE0); gsrputc(d2); }
      else
      {
        if(cur_a_adr == d1) { gsrputc(0xF9); gsrputc(d2); }
        else                { gsrputc(0xF7); gsrputc(d1); gsrputc(d2); cur_a_adr = d1; }
      }
      break;
    case GSRDUMP_YM2612_PORT2:
      if(cur_b_adr == d1) { gsrputc(0xFA); gsrputc(d2); }
      else                { gsrputc(0xF8); gsrputc(d1); gsrputc(d2); cur_b_adr = d1; }
      break;
    case GSRDUMP_PSG:
      gsrputc(0xFB); gsrputc(d1);
      break;
  }
}


#define OPEN()      ( Compress_GSR ? (FILE *)gzdopen(gsr_file, "wb") : fdopen(gsr_file, "wbS") )
#define CLOSE(file) {                                      \
  if(Compress_GSR) { FLUSH(file); gzclose((GZFILE)file); } \
  else             { fclose(file); } }
#define WRITE(file, data, size) \
  ( Compress_GSR ? gzwrite((GZFILE)file, data, size) == size : fwrite(data, size, 1, file) == 1 )
#define FLUSH(file) if(buf_used) { \
  gzwrite((GZFILE)file, State_Buffer, buf_used); buf_used = 0; }
#define PUTC(file, c) {                      \
  if(Compress_GSR)                           \
  { if(buf_used >= 0x10000) { FLUSH(file); } \
    State_Buffer[buf_used++] = c; }          \
  else { putc(c, file); } }

void __stdcall Stop_GSR_Dump(void)
{
  FILE     *fout;
  GSRHEADER gsrhdr;
  int       buf_used, i, j, c, diff, len, total;
  int       sample_matching_marge, size_marge;
  LPINT     replace_sample = NULL;
  const int rate = CPU_Mode ? 312 * 50 : 262 * 60;

  if(!GSR_Dumping) return;

#ifdef GSR_DAC_LOG
  if(fDACLog) { fclose(fDACLog); fDACLog = NULL; }
#endif

  flush_dac_data();

  fout = OPEN();
  if(!fout)
  { _close(gsr_file); fclose(fdata);
    sample_buf.quit(); samples_table.quit();
    GSR_Dumping = 0; return; }

  buf_used = 0; // para compactao

  total = samples_table.ndata;
  replace_sample = (LPINT)malloc(total * 4);
  if(!replace_sample) goto __end;
  for(i = 0; i < total; i++) replace_sample[i] = i;

  // comparando os samples na tabela
  sample_matching_marge = 21 - GSR_DAC_Precision;
  if     (GSR_DAC_Precision == 20) size_marge = 0;
  else if(GSR_DAC_Precision >= 16) size_marge = 3;
  else if(GSR_DAC_Precision >= 12) size_marge = 5;
  else                             size_marge = 7;

  for(i = 0; i < samples_table.ndata - 1; i++)
  {
    if(samples_table.data[i].size > 0) {

    for(j = i + 1; j < samples_table.ndata; j++)
    {
      if(samples_table.data[j].size > 0) {

      // verificando o tamanho
      if(samples_table.data[i].size >= samples_table.data[j].size - size_marge &&
         samples_table.data[i].size <= samples_table.data[j].size + size_marge)
      {
        if(samples_table.data[i].size > samples_table.data[j].size)
          len = samples_table.data[j].size;
        else
          len = samples_table.data[i].size;

        for(diff = c = 0; c < len; c++)
          diff += abs((int)samples_table.data[i].ptr[c] - (int)samples_table.data[j].ptr[c]);

        diff /= len;
        if(diff < sample_matching_marge) // sample igual!!
        {
          if(samples_table.data[j].size > samples_table.data[i].size)
          { samples_table.swap_data(i, j); true_samples_table.swap_data(i, j); }
          // deletando os samples que no sero mais usados
          samples_table.delete_data(j);
          true_samples_table.delete_data(j);
          replace_sample[j] = i;
        }
      } }

    } }
  }

  for(i = 0; i < samples_table.ndata; i++)
  {
    if(samples_table.data[i].size <= 0)
    {
      if(i < samples_table.ndata--)
      {
        memcpy(samples_table.data + i, samples_table.data + i + 1,
               (samples_table.ndata - i) * sizeof(SAMPLE_DATA));
        memcpy(true_samples_table.data + i, true_samples_table.data + i + 1,
               (true_samples_table.ndata - i) * sizeof(SAMPLE_DATA));
        for(j = 0; j < total; j++)
        { if(replace_sample[j] >= i) replace_sample[j]--; }
        i--;
      }
    }
  }

  memset(&gsrhdr, 0, sizeof(GSRHEADER));
  gsrhdr.dwType         = GSR_TYPE;
  gsrhdr.dwVersion      = GSR_VERSION;
  gsrhdr.dwTableSamples = samples_table.ndata;
  gsrhdr.dwTotalLength  = n_frames;
  if(CPU_Mode) gsrhdr.dwFlags |= GSR_FLAG_PAL_SYSTEM;
  // copiando os dados da rom
  memcpy(&gsrhdr.griRomInfo, &Game->Genesis, sizeof(GENESIS_ROM_INFO));
  
  if(!WRITE(fout, &gsrhdr, sizeof(GSRHEADER))) goto __end;
  if(!WRITE(fout, &ci, sizeof(CHIPSINFO)))     goto __end;

  for(i = 0; i < samples_table.ndata; i++)
  {
    if(!WRITE(fout, &(true_samples_table.data[i].size), sizeof(int))) goto __end;
    // sample rate: reservado para verses futuras
    if(!WRITE(fout, &(rate), sizeof(int)))                            goto __end;
    if(!WRITE(fout, true_samples_table.data[i].ptr, true_samples_table.data[i].size))
      goto __end;
  }

  fseek(fdata, 0, SEEK_SET);
  while((c = getc(fdata)) != EOF)
  {
    if((c & 0xF0) == 0xE0)
    { PUTC(fout, c);
      c = getc(fdata);
      PUTC(fout, c); }
    else switch(c)
    {
      case 0xFC:
        i = getc(fdata);
        j = getc(fdata);
        j |= (i << 8);
        PUTC(fout, c);
        PUTC(fout, replace_sample[j] >> 8);
        PUTC(fout, replace_sample[j] & 0xFF);
        i = getc(fdata);
        j = getc(fdata);
        PUTC(fout, i);
        PUTC(fout, j);
        break;
      case 0xF7:
      case 0xF8:
        PUTC(fout, c);
        c = getc(fdata);
      case 0xF9:
      case 0xFA:
      case 0xFB:
        PUTC(fout, c);
        c = getc(fdata);
      case 0xF0:
        PUTC(fout, c);
        break;
    }
  }
  PUTC(fout, 0xFF);

__end:
  CLOSE(fout);
  fclose(fdata);
  if(replace_sample) free(replace_sample);

  sample_buf.quit();
  samples_table.quit();
  true_sample_buf.quit();
  true_samples_table.quit();

  GSR_Dumping = 0;
}
