
#define DIRECTDRAW_VERSION 0x0700
#define DIRECT3D_VERSION   0x0700

#include <ddraw.h>
#include <d3d.h>

#include "gmain.h"
#include "ggraphic.h"
#include "vdp_io.h"
#include "vdp_rend.h"
#include "misc.h"
#include "blit.h"
#include "sms.h"


extern LPDIRECTDRAW         lpDD_Init;
extern LPDIRECTDRAW7        lpDD;
extern LPDIRECTDRAWSURFACE7 lpDDS_Primary;
extern LPDIRECTDRAWSURFACE7 lpDDS_Flip;
extern LPDIRECTDRAWSURFACE7 lpDDS_Back;
extern LPDIRECTDRAWSURFACE7 lpDDS_Blit;
extern LPDIRECTDRAWCLIPPER  lpDDC_Clipper;

LPDIRECT3D7                 lpD3D         = NULL;
LPDIRECT3DDEVICE7           lpD3D_Device  = NULL;
LPDIRECTDRAWSURFACE7        lpDDS_Texture = NULL;
D3DVIEWPORT7                d3dvp;
D3DVERTEX                   lpvPlane[4];
D3DMATRIX                   mWorld, mView, mProj;


HRESULT WINAPI Enum_Textures(LPDDPIXELFORMAT pddpf, LPVOID pret)
{
  if(pddpf->dwFlags & (DDPF_LUMINANCE | DDPF_BUMPLUMINANCE | DDPF_BUMPDUDV))
    return D3DENUMRET_OK;

  if(pddpf->dwFourCC != 0) return D3DENUMRET_OK;

  if(pddpf->dwFlags & DDPF_ALPHAPIXELS) return D3DENUMRET_OK;

  if(pddpf->dwRGBBitCount != 16) return D3DENUMRET_OK;

  memcpy(pret, pddpf, sizeof(DDPIXELFORMAT));

  if(pddpf->dwGBitMask == 0x03E0)
  {
    if(Mode_555 != 0) return D3DENUMRET_CANCEL;
  }
  else
  {
    if(Mode_555 != 1) return D3DENUMRET_CANCEL;
  }

  return D3DENUMRET_OK;
}

void D3D_Fail(HWND hWnd, char *err)
{
  End_D3D();
  MessageBox(hWnd, err, "Error Direct3D", MB_ICONSTOP | MB_OK);
  DestroyWindow(hWnd);
  exit(0);
}

int __stdcall Init_D3D(HWND hWnd)
{
  int                 rend;
  HRESULT             rval;
  DDSURFACEDESC2      ddsd;
  DDPIXELFORMAT       ddpf;
  int                 rx, ry, filter;
  D3DMATERIAL7        ddMaterial;
  D3DDEVICEDESC7      ddDesc;
  D3DTEXTUREMAGFILTER ddMagFilter;
  D3DTEXTUREMINFILTER ddMinFilter;

  if(DirectDrawCreateEx(0, (LPVOID *)&lpDD_Init, IID_IDirectDraw7, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: DirectDrawCreate.\nCheck your DirectX version."); return 0; }
  if(lpDD_Init->QueryInterface(IID_IDirectDraw7, (LPVOID *)&lpDD) != DD_OK)
  { D3D_Fail(hWnd, "Error: QueryInterface.\nCheck your DirectX version."); return 0; }
  if(lpDD_Init->QueryInterface(IID_IDirect3D7, (LPVOID *)&lpD3D) != DD_OK)
  { D3D_Fail(hWnd, "Error: QueryInterface.\nCheck your DirectX version."); return 0; }

  lpDD_Init->Release();
  lpDD_Init = NULL;

  if(Full_Screen) rend = Render_FS;
  else            rend = Render_W;

  if(Full_Screen)
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_FPUSETUP);
  else
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL | DDSCL_FPUSETUP);

  if(rval != DD_OK)
  { D3D_Fail(hWnd, "Error: SetCooperativeLevel.\nCheck your DirectX version."); return 0; }

  if(Full_Screen)
  {
    if(Res_X < 640 || Res_Y < 480)
    { rend = 0;
      if(Have_MMX) Blit_FS = Blit_X1_MMX;
      else         Blit_FS = Blit_X1; }

    if(!rend)
    { rx = 320;  ry = 240; }
    else
    { rx = Res_X;  ry = Res_Y; }
    if(lpDD->SetDisplayMode(rx, ry, 16, 0, 0) != DD_OK)
    { D3D_Fail(hWnd, "Error: SetDisplayMode."); return 0; }

    if((Res_X < (320 * 3) || Res_Y < (240 * 3)) && Render_FS > 14) 
    {
      Render_FS = 1;
      if(Have_MMX) Blit_FS = Blit_X2_MMX;
      else         Blit_FS = Blit_X2;
    }
  }

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize   = sizeof(DDSURFACEDESC2);

  if(Full_Screen && VSync_FS)
  {
    ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 2;
  }
  else
  {
    ddsd.dwFlags        = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
  }

  if(lpDD->CreateSurface(&ddsd, &lpDDS_Primary, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  if(Full_Screen)
  {
    if(VSync_FS)
    {
      ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
      if(lpDDS_Primary->GetAttachedSurface(&ddsd.ddsCaps, &lpDDS_Flip) != DD_OK)
      { D3D_Fail(hWnd, "Error: GetAttachedSurface."); return 0; }
      lpDDS_Blit = lpDDS_Flip;
    }
    else lpDDS_Blit = lpDDS_Primary;
  }
  else
  {
    if(lpDD->CreateClipper(0, &lpDDC_Clipper, NULL ) != DD_OK)
    { D3D_Fail(hWnd, "Error: CreateClipper."); return 0; }

    if(lpDDC_Clipper->SetHWnd(0, hWnd) != DD_OK)
    { D3D_Fail(hWnd, "Error: SetghWnd."); return 0; }

    if(lpDDS_Primary->SetClipper(lpDDC_Clipper) != DD_OK)
    { D3D_Fail(hWnd, "Error: SetClipper."); return 0; }
  }


  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize   = sizeof(DDSURFACEDESC2);
  ddsd.dwFlags  = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  if(Full_Screen) { ddsd.dwWidth = Res_X; ddsd.dwHeight = Res_Y; }
  else
  { ddsd.dwWidth  = GetSystemMetrics(SM_CXFULLSCREEN);
    ddsd.dwHeight = GetSystemMetrics(SM_CYFULLSCREEN); }
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY | DDSCAPS_3DDEVICE;
  if(lpDD->CreateSurface(&ddsd, &lpDDS_Back, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  // iniciando o direct3D
  if(lpD3D->CreateDevice(IID_IDirect3DHALDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
  {
    if(lpD3D->CreateDevice(IID_IDirect3DMMXDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
    {
      if(lpD3D->CreateDevice(IID_IDirect3DRefDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
      { D3D_Fail(hWnd, "Error: CreateDevice."); return 0; }
    }
  }

  if(Force_Mode == 2)
  {
    memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
    ddsd.dwSize = sizeof(DDSURFACEDESC2);
    lpDD->GetDisplayMode(&ddsd);
    if(ddsd.ddpfPixelFormat.dwRGBBitCount == 16)
    {
      if(ddsd.ddpfPixelFormat.dwGBitMask == 0x03E0) Mode_555 = 1;
      else                                          Mode_555 = 0;
    }
    else Mode_555 = Force_Mode;
  }
  else Mode_555 = Force_Mode;

  lpD3D_Device->EnumTextureFormats(Enum_Textures, &ddpf);
  Mode_555 = ddpf.dwGBitMask == 0x03E0;

  Recalculate_Palettes();

  // viewport
  d3dvp.dwX = d3dvp.dwY = 0;
  if(Full_Screen) { d3dvp.dwWidth = Res_X; d3dvp.dwHeight = Res_Y; }
  else
  { d3dvp.dwWidth  = GetSystemMetrics(SM_CXFULLSCREEN);
    d3dvp.dwHeight = GetSystemMetrics(SM_CYFULLSCREEN); }
  d3dvp.dvMinZ = 0.0f;
  d3dvp.dvMaxZ = 1.0f;
  if(lpD3D_Device->SetViewport(&d3dvp) != DD_OK)
  { D3D_Fail(hWnd, "Error: SetViewport."); return 0; }

  // plano onde a imagem  impressa
  memset(lpvPlane, 0, sizeof(D3DVERTEX) * 4);
  lpvPlane[0].x = 1.0f;  lpvPlane[0].y = -1.0f; lpvPlane[0].z = 0.0f; lpvPlane[0].nz = 1.0f;
  lpvPlane[1].x = -1.0f; lpvPlane[1].y = -1.0f; lpvPlane[1].z = 0.0f; lpvPlane[1].nz = 1.0f;
  lpvPlane[2].x = 1.0f;  lpvPlane[2].y = 1.0f;  lpvPlane[2].z = 0.0f; lpvPlane[2].nz = 1.0f;
  lpvPlane[3].x = -1.0f; lpvPlane[3].y = 1.0f;  lpvPlane[3].z = 0.0f; lpvPlane[3].nz = 1.0f;
  // coordenadas da textura (imagem do emulador)
  lpvPlane[0].tu = 1.0f; lpvPlane[0].tv = 1.0f;
  lpvPlane[1].tu = 0.0f; lpvPlane[1].tv = 1.0f;
  lpvPlane[2].tu = 1.0f; lpvPlane[2].tv = 0.0f;
  lpvPlane[3].tu = 0.0f; lpvPlane[3].tv = 0.0f;

  memset(&ddMaterial, 0, sizeof(D3DMATERIAL7));
  ddMaterial.ambient.r = 1.0f;
  ddMaterial.ambient.g = 1.0f;
  ddMaterial.ambient.b = 1.0f;
  lpD3D_Device->SetMaterial(&ddMaterial);

  lpD3D_Device->SetRenderState(D3DRENDERSTATE_AMBIENT, 0x00FFFFFF);

  mWorld._11 = mWorld._22 = mWorld._33 = mWorld._44 = 1.0f;
  mWorld._12 = mWorld._13 = mWorld._14 = mWorld._41 = 0.0f;
  mWorld._21 = mWorld._23 = mWorld._24 = mWorld._42 = 0.0f;
  mWorld._31 = mWorld._32 = mWorld._34 = mWorld._43 = 0.0f;
  lpD3D_Device->SetTransform(D3DTRANSFORMSTATE_WORLD, &mWorld);
  mView = mWorld;
  mView._43 = 2.0f;
  lpD3D_Device->SetTransform(D3DTRANSFORMSTATE_VIEW, &mView);
  mProj = mWorld;
  mProj._11 =  2.0f;
  mProj._22 =  2.0f;
  mProj._34 =  1.0f;
  mProj._43 = -1.0f;
  mProj._44 =  0.0f;
  lpD3D_Device->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &mProj);

  lpD3D_Device->GetCaps(&ddDesc);

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize          = sizeof(DDSURFACEDESC2);
  ddsd.dwFlags         = DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_HEIGHT | DDSD_WIDTH;
  ddsd.ddpfPixelFormat = ddpf;

  if(ddDesc.deviceGUID == IID_IDirect3DHALDevice)
  { ddsd.ddsCaps.dwCaps  = DDSCAPS_TEXTURE;
    ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE; }
  else
  { ddsd.ddsCaps.dwCaps  = DDSCAPS_TEXTURE | DDSCAPS_SYSTEMMEMORY; }

  if(rend > 13)      { rx = 960; ry = 720; }
  else if(rend > 0)  { rx = 640; ry = 480; }
	else               { rx = 320; ry = 240; }

  if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2)
  {
    for(ddsd.dwWidth  = 1; rx > (int)ddsd.dwWidth;  ddsd.dwWidth <<= 1);
    for(ddsd.dwHeight = 1; ry > (int)ddsd.dwHeight; ddsd.dwHeight <<= 1);
  }
  if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY)
  {
    if(ddsd.dwWidth > ddsd.dwHeight) ddsd.dwHeight = ddsd.dwWidth;
    else                             ddsd.dwWidth  = ddsd.dwHeight;
  }

  // textura onde o emulador imprime a imagem
  if(lpDD->CreateSurface(&ddsd, &lpDDS_Texture, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  lpD3D_Device->SetTexture(0, lpDDS_Texture);
  if(Full_Screen) filter = D3D_Filter_FS;
  else            filter = D3D_Filter_W;

  switch(filter)
  {
    case 0: ddMagFilter = D3DTFG_POINT;         ddMinFilter = D3DTFN_POINT;       break;
    case 1: ddMagFilter = D3DTFG_LINEAR;        ddMinFilter = D3DTFN_LINEAR;      break;
    case 2: ddMagFilter = D3DTFG_FLATCUBIC;     ddMinFilter = D3DTFN_LINEAR;      break;
    case 3: ddMagFilter = D3DTFG_GAUSSIANCUBIC; ddMinFilter = D3DTFN_LINEAR;      break;
    case 4: ddMagFilter = D3DTFG_ANISOTROPIC;   ddMinFilter = D3DTFN_ANISOTROPIC; break;
  }
  lpD3D_Device->SetTextureStageState(0, D3DTSS_MINFILTER, ddMinFilter);
  lpD3D_Device->SetTextureStageState(0, D3DTSS_MAGFILTER, ddMagFilter);

  return 1;
}

void __stdcall End_D3D(void)
{
  if(lpDDC_Clipper) { lpDDC_Clipper->Release(); lpDDC_Clipper = NULL; }
  if(lpDDS_Back)    { lpDDS_Back->Release();    lpDDS_Back = NULL; }
  if(lpDDS_Texture) { lpDDS_Texture->Release(); lpDDS_Texture = NULL; }
  if(lpDDS_Flip)    { lpDDS_Flip->Release();    lpDDS_Flip = NULL; }
  if(lpDDS_Primary) { lpDDS_Primary->Release(); lpDDS_Primary = NULL; }
  if(lpD3D_Device)  { lpD3D_Device->Release();  lpD3D_Device = NULL; }
  if(lpD3D)         { lpD3D->Release();         lpD3D = NULL; }
  if(lpDD)
  { lpDD->SetCooperativeLevel(ghWnd, DDSCL_NORMAL);
    lpDD->Release();
    lpDD = NULL; }
  lpDDS_Blit = NULL;
}


int __stdcall Clear_D3D_Primary_Screen(HWND hWnd)
{
  return Clear_DDraw_Primary_Screen(hWnd);
}


int __stdcall Clear_D3D_Back_Screen(HWND hWnd)
{
  DDSURFACEDESC2 ddsd;
  DDBLTFX        ddbltfx;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);
  memset(&ddbltfx, 0, sizeof(DDBLTFX));
  ddbltfx.dwSize      = sizeof(DDBLTFX);
  ddbltfx.dwFillColor = 0;
  lpDDS_Back->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);
  memset(&ddbltfx, 0, sizeof(DDBLTFX));
  ddbltfx.dwSize      = sizeof(DDBLTFX);
  ddbltfx.dwFillColor = 0;
  lpDDS_Texture->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);

  return 1;
}


void __stdcall Restore_D3D_Primary(void)
{
  Restore_DDraw_Primary();
}


void __stdcall Flip_D3D_FS(HWND hWnd)
{
  RECT           RectSrc, RectDest;
  int            sx, sy, rx, ry;
  float          fx, fy;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 40; }

    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 32; }

    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  if     (Render_FS > 13) { rx = sx * 3;  ry = sy * 3; }
  else if(Render_FS > 0)  { rx = sx << 1; ry = sy << 1; }
  else                    { rx = sx;      ry = sy; }

  RectDest.left   = RectDest.top = 0;
  RectDest.right  = Res_X;
  RectDest.bottom = Res_Y;
  RectSrc         = RectDest;

  if(Stretch_FS)
  {
    fx = 1.0f;
    fy = 1.0f;
  }
  else
  {
    fx = ((float)(rx >> 1) / (float)(d3dvp.dwWidth >> 1)) + 0.0001f;
    fy = ((float)(ry >> 1) / (float)(d3dvp.dwHeight >> 1)) + 0.0001f;
  }

  lpvPlane[0].x = fx;  lpvPlane[0].y = -fy;
  lpvPlane[1].x = -fx; lpvPlane[1].y = -fy;
  lpvPlane[2].x = fx;  lpvPlane[2].y = fy;
  lpvPlane[3].x = -fx; lpvPlane[3].y = fy;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);
  rval = lpDDS_Texture->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
  if(rval != DD_OK) return;
  Blit_FS((LPBYTE)ddsd.lpSurface, ddsd.lPitch, sx, sy, (336 - sx) << 1);
  fx = (float)rx / (float)ddsd.dwWidth;
  fy = (float)ry / (float)ddsd.dwHeight;
  lpvPlane[0].tu = fx;   lpvPlane[0].tv = fy;
  lpvPlane[1].tu = 0.0f; lpvPlane[1].tv = fy;
  lpvPlane[2].tu = fx;   lpvPlane[2].tv = 0.0f;
  lpvPlane[3].tu = 0.0f; lpvPlane[3].tv = 0.0f;
  lpDDS_Texture->Unlock(NULL);

  if(lpD3D_Device->BeginScene() == DD_OK)
  {
    lpD3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_VERTEX, lpvPlane, 4, NULL);
    lpD3D_Device->EndScene();
  }

  if(VSync_FS)
  { lpDDS_Flip->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
    lpDDS_Primary->Flip(NULL, DDFLIP_WAIT); }
  else
  { lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL); }
}

HRESULT __stdcall RestoreGraphics(void);

void __stdcall Flip_D3D_W(HWND hWnd)
{
  RECT           RectSrc, RectDest;
  int            sx, sy, rx, ry;
  float          fx, fy;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;
  POINT          p;

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 40; }

    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 32; }

    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  p.x = p.y = 0;
  GetClientRect(hWnd, &RectDest);
  ClientToScreen(hWnd, &p);

  if     (Render_W > 13) { rx = sx * 3;  ry = sy * 3; }
  else if(Render_W > 0)  { rx = sx << 1; ry = sy << 1; }
  else                   { rx = sx;      ry = sy; }

  RectSrc.right  = d3dvp.dwWidth;
  RectSrc.bottom = d3dvp.dwHeight;
  if(RectSrc.right > RectDest.right)
  { RectSrc.left  = (RectSrc.right >> 1) - (RectDest.right >> 1);
    RectSrc.right = RectDest.right + RectSrc.left; }
  else
  { RectSrc.left   = 0;
    RectDest.right = RectSrc.right; }
  if(RectSrc.bottom > RectDest.bottom)
  { RectSrc.top     = (RectSrc.bottom >> 1) - (RectDest.bottom >> 1);
    RectSrc.bottom  = RectDest.bottom + RectSrc.top; }
  else
  { RectSrc.top     = 0;
    RectDest.bottom = RectSrc.bottom; }

  if(Stretch_W)
  {
    fx = ((float)(RectDest.right >> 1) / (float)(d3dvp.dwWidth >> 1));
    fy = ((float)(RectDest.bottom >> 1) / (float)(d3dvp.dwHeight >> 1));
  }
  else
  {
    fx = ((float)(rx >> 1) / (float)(d3dvp.dwWidth >> 1));
    fy = ((float)(ry >> 1) / (float)(d3dvp.dwHeight >> 1));
  }

  lpvPlane[0].x = fx - 0.00005f;  lpvPlane[0].y = -fy + 0.00005f;
  lpvPlane[1].x = -fx;            lpvPlane[1].y = -fy + 0.00005f;
  lpvPlane[2].x = fx - 0.00005f;  lpvPlane[2].y = fy;
  lpvPlane[3].x = -fx;            lpvPlane[3].y = fy;

  RectDest.left    = p.x;  RectDest.top     = p.y;
  RectDest.right  += p.x;  RectDest.bottom += p.y;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);
  rval = lpDDS_Texture->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
  if(rval != DD_OK) return;
  Blit_W((LPBYTE)ddsd.lpSurface, ddsd.lPitch, sx, sy, (336 - sx) << 1);
  fx = (float)rx / (float)ddsd.dwWidth;
  fy = (float)ry / (float)ddsd.dwHeight;
  lpvPlane[0].tu = fx;   lpvPlane[0].tv = fy;
  lpvPlane[1].tu = 0.0f; lpvPlane[1].tv = fy;
  lpvPlane[2].tu = fx;   lpvPlane[2].tv = 0.0f;
  lpvPlane[3].tu = 0.0f; lpvPlane[3].tv = 0.0f;
  lpDDS_Texture->Unlock(NULL);

  if(lpD3D_Device->BeginScene() == DD_OK)
  {
    lpD3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_VERTEX, lpvPlane, 4, NULL);
    lpD3D_Device->EndScene();
  }

  if(VSync_W)
  {
    int vb;
    lpDD->GetVerticalBlankStatus(&vb);
    if(!vb) lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);
  }
  rval = lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
  if(rval == DDERR_SURFACELOST) rval = RestoreGraphics();
}
